import boto3
import io
import json 
from typing import Tuple

class S3:
    client: boto3.client

    def __init__(self) -> None:
        self.client = boto3.client('s3')

    def _streaming_body_to_dict(self, payload):
        file_like_obj = io.BytesIO(payload.read())
        response = json.loads(file_like_obj.getvalue())
        return response
    
    def put_json_object(self, bucket_name: str, file_name: str, object: dict):
        return self.client.put_object(
            Body = json.dumps(object),
            Bucket = bucket_name,
            Key = file_name
        )
    
    def get_object(self, bucket_name: str, file_name: str):
        response = self.client.get_object(
            Bucket = bucket_name,
            Key = file_name            
        )
        return self._streaming_body_to_dict(response["Body"])
    
    def get_presigned_url(self, bucket_name: str, file_name: str, expires_in: int = 3600):
        return self.client.generate_presigned_url(
            'get_object',
            Params = {
                "Bucket": bucket_name,
                "Key": file_name,
            },
            ExpiresIn = expires_in
        )
    
    def get_s3_location_from_bucket_file(bucket_name: str, file_name: str) -> str:
        return f"{bucket_name}/{file_name}"
    
    def get_bucket_file_from_s3_location(s3_location: str) -> Tuple[str, str]:
        bucket, file_name = s3_location.split('/')[0], "/".join(s3_location.split('/')[1:])
        return bucket, file_name
    


# import boto3
# from authlib.jose import jwt
# import requests
# import json

# class Cognito:
#     def __init__(self, pool_id: str, client_id: str) -> None:
#         self.pool_id = pool_id
#         self.client_id = client_id
#         self.client = boto3.client('cognito-idp')
#         self.cognito_pool_domain_url = None
#         self.cognito_known_tokens_url = None

#     def delete_user(self, username: str):
#         try:
#             response = self.client.admin_delete_user(
#             UserPoolId=self.pool_id,
#             Username=username
#             )
#             return response
#         except Exception as e:
#             raise Exception(f"Error {e}: Failed to delete User {username}")

#     def create_user(self, user: User):
#         try:
#             response = self.client.admin_create_user(
#                 UserPoolId= self.pool_id,
#                 Username = user.email,
#                 UserAttributes=[
#                     {
#                         'Name': 'Given Name',
#                         'Value': user.first,
#                         'Name': 'Family Name',
#                         'Value': user.last,
#                         'Name': 'Full Name',
#                         'Value': user.fullname,
#                         'Name': 'name',
#                         'Value': user.fullname,
#                         'Name': 'nickname',
#                         'Value': user.nickname,
#                         'Name': 'email',
#                         'Value': user.email
#                     },
#                 ],
#                 ValidationData=[
#                      {
#                         'Name': 'Given Name',
#                         'Value': user.first,
#                         'Name': 'Family Name',
#                         'Value': user.last,
#                         'Name': 'Full Name',
#                         'Value': user.fullname,
#                         'Name': 'name',
#                         'Value': user.fullname,                        
#                         'Name': 'nickname',
#                         'Value': user.nickname,
#                         'Name': 'email',
#                         'Value': user.email
#                     },
#                 ],
#                 TemporaryPassword=user.password, #should come from user
#                 ForceAliasCreation=False,
#                 MessageAction='SUPPRESS',
#                 DesiredDeliveryMediums=[
#                     'EMAIL',
#                 ],
#                 ClientMetadata={
#                     'string': 'string'
#                 }
#             )
#             return response
#         except Exception as e:
#             raise Exception(f"Error {e}: Could not create user {user.fullname}")

#     def _init_auth(self, cognito_username: str, user: User):
#         """Initiate Auth (SRP) -> Returns Challenge"""
#         try:
#             response = self.client.initiate_auth(            
#                 AuthFlow='USER_PASSWORD_AUTH', #'USER_SRP_AUTH'|'REFRESH_TOKEN_AUTH'|'REFRESH_TOKEN'|'CUSTOM_AUTH'|'ADMIN_NO_SRP_AUTH'| 'USER_PASSWORD_AUTH',
#                 AuthParameters={
#                     'USERNAME': user.email,
#                     'name': user.fullname,
#                     'email':user.email,
#                     'PASSWORD': user.password
#                 },
#                 ClientId=self.client_id,
#             )
#             return response
#         except Exception as e:
#             raise Exception(f"Error {e}: Issue during initialising auth for user {cognito_username}")

#     def _respond_to_auth_challenge(self, init_auth_response, user: User, cognito_username: str): 
#         """Respond to Auth Challenge -> Cognito Tokens"""
#         try:
#             cognito_username = init_auth_response["ChallengeParameters"].get('USER_ID_FOR_SRP') #needs testing
#             cn = init_auth_response["ChallengeName"]
#             sess = init_auth_response["Session"]
#             challenge_params = init_auth_response["ChallengeParameters"]
#             challenge_resp = {
#                     'USERNAME': user.email,
#                     'userAttributes.name': user.fullname,
#                     'NEW_PASSWORD':user.password
#                 }
#             response = self.client.respond_to_auth_challenge(
#                 ClientId=self.client_id,
#                 ChallengeName=cn,
#                 Session=sess,
#                 ChallengeResponses=challenge_resp
#             )
#             return response
#         except Exception as e:
#              raise Exception(f"Error {e}: Issue responding to auth challenge for user {user.email}")
        
#     # def get_token(self, user: User, cognito_username: str):
#     #     init_auth_response = self._init_auth(cognito_username, user)
#     #     response = self._respond_to_auth_challenge(init_auth_response, user, cognito_username)
#     #     return response

#     # def sign_up(self, user: User): 
#     #     response = self.create_user(user)
#     #     cognito_username = response['User'].get('Username')
#     #     get_token_response = self.get_token(user, cognito_username)
#     #     return get_token_response

#     # def get_JWT_from_code(self, code):
#     #     try:
#     #         payload=f'grant_type=authorization_code&client_id={self.client_id}&code={code}'#&redirect_uri={call_back_uri}
#     #         headers = {
#     #         'Content-Type': 'application/x-www-form-urlencoded',
#     #         }
#     #         response = requests.request("POST", self.cognito_pool_domain, headers=headers, data=payload)
#     #         tokens = json.loads(response.text)
#     #         return tokens
#     #     except Exception as e:
#     #         raise Exception(f"Error {e}: Couldn't get JWT from auth code")

#     # def code_to_jwt_dict(self, code):
#     #     tokens = self.code_to_JWT(code)
#     #     refresh_token = tokens['refresh_token']
#     #     access_token = tokens['access_token']
#     #     JWT = tokens['id_token']
#     #     response = requests.request("GET", self.cognito_known_tokens_url, headers={}, data={})
#     #     public = json.loads(response.text) #loads a json str into dict
#     #     return jwt.decode(JWT, public) #decode jwt 



from typing import Optional, Dict, Any
import json
from my_aws_helpers.errors import *

class API:
    def response(code: int, body: Optional[str] = None):
        return {
            "statusCode": code, 
            "headers": {"Access-Control-Allow-Origin": "*"},
            "body": body,
        }

    def parse_payload(event: Dict[str, Any]):
        payload = {}
        if event.get("queryParameters"): payload["queryParameters"] = event["queryParameters"]
        if event.get("pathParameters"): payload["pathParameters"] = event["pathParameters"]
        if event.get("body"): payload["body"] = json.loads(event["body"])
        return payload
    
    def handle_error_response(func):
        def wrapper(event, context):
            try:
                response = func(event, context)        
                return API.response(
                    code = 200, 
                    body = json.dumps(response) 
                )
            except ClientError as e:
                return API.response(
                    code = 400, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except NotFoundError as e:
                return API.response(
                    code = 404, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except ServerError as e:
                return API.response(
                    code = 500, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except Exception as e:
                return API.response(
                    code = 500, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
        return wrapper



import boto3
from typing import Optional
import json
from uuid import uuid4
from my_aws_helpers.api import API
from my_aws_helpers.errors import *

class SFN:
    client: boto3.client

    def __init__(self) -> None:
        self.client = boto3.client('stepfunctions')

    def start_execution(self, sfn_arn: str, event: Optional[dict] = None, name: Optional[str] = uuid4().hex):
        input_event = json.dumps(event) if event else event
        return self.client.start_execution(
            stateMachineArn = sfn_arn,
            name = name,
            input = input_event,
        )
    
    def handle_error_response(func):
        def wrapper(event, context):
            try:
                response = func(event, context)        
                return API.response(
                    code = 200, 
                    body = json.dumps(response) 
                )
            except AlreadyExists as e:
                return API.response(
                    code = 201, 
                    body = json.dumps({"Success": f"{e}"}) 
                )                
            except ClientError as e:
                return API.response(
                    code = 400, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except NotFoundError as e:
                return API.response(
                    code = 404, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except ServerError as e:
                return API.response(
                    code = 500, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
            except Exception as e:
                return API.response(
                    code = 500, 
                    body = json.dumps({"Error": f"{e}"}) 
                )
        return wrapper    
    




from typing import List, Any, Optional
from datetime import datetime, date
import boto3
from abc import ABC, abstractclassmethod, abstractmethod

class MetaData:
    """
    This class is a convenience class, 
    each of its attributes will be attached to objects that inherit from `BaseTableObject`
    """
    created_by: Optional[str]
    created_on: Optional[datetime]
    updated_by: Optional[str]
    updated_on: Optional[datetime]

    def set_timestamp(self, ts: Any) -> datetime:
        """Be absolutely sure timestamps are datetimes"""
        if isinstance(ts, datetime):
            return ts
        else:
            return datetime.now()

    def __init__(self, **kwargs) -> None:
        self.created_by = kwargs["created_by"] if kwargs.get("created_by") else self._get_user()
        self.updated_by = kwargs["updated_by"] if kwargs.get("updated_by") else self._get_user()
        self.created_on = self.set_timestamp(ts=kwargs.get("created_on"))
        self.updated_on = self.set_timestamp(ts=kwargs.get("updated_on"))

    def _get_user(self):
        """This should probably do some clever thing to get the actual user details from the token or something"""
        return ""


class BaseTableObject(MetaData):
    """
    An Abstract class that helps ensure your objects
    conform to the AssetTable schema and
    implement serialisation/deserialisation for Dynamo
    """
    pk: str
    sk: str

    def _get_pk(self):
        pass

    def _get_sk(self):
        pass

    @abstractclassmethod
    def _from_dynamo_representation():
        """
        Deserialises this object from Dynamo Representation
        """
        pass

    @abstractmethod
    def _to_dynamo_representation():
        """
        Serialises this object to Dynamo Representation
        """        
        pass

    def _optional_get(self, kwargs: dict, key: str, default: Any):
        return kwargs.get(key) if kwargs.get(key) else default
    
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.pk = self._optional_get(kwargs=kwargs, key='pk', default=self._get_pk())
        self.sk = self._optional_get(kwargs=kwargs, key='sk', default=self._get_sk())

class BaseQueries(ABC):
    table_name: str

    def __init__(self, table_name: str) -> None:
        self.table_name = table_name

class Dynamo:
    def __init__(self, table_name: str) -> None:
        ddb = boto3.resource('dynamodb')
        self.table = ddb.Table(table_name)
    
    def put_item(self, item: dict):
        return self.table.put_item(Item=item)
    
    def get_item(self, item: dict):
        return self.table.get_item(Item=item)
    
    def delete_item(self, item: dict):
        return self.table.delete_item(Item=item)
    
    def batch_put(self, items: List[dict]) -> None:
        with self.table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)
        return 
    
    def batch_delete(self, items: List[dict]) -> None:
        with self.table.batch_writer() as batch:
            for item in items:
                batch.delete_item(Item=item)
        return

    def to_dynamo_representation(obj: dict):
        """
        Attempts to put common datatype transformations in one spot
        """
        new_obj = dict()
        for key, value in obj.items():
            new_obj[key] = _datatype_map(value=value)
        return new_obj
    
def _datatype_map(value: Any):
    if (isinstance(value, date)) or (isinstance(value, datetime)):
        return value.isoformat()
    if (isinstance(value, list)):
        return [_datatype_map(value = item) for item in value]
    if (isinstance(value, dict)):
        new_obj = dict()
        for k, v in value.items():
            new_obj[k] = _datatype_map(value=v)
        return new_obj
    return value